"""
Accessory methods for Connect
"""

class ParameterIsMissingException(Exception):
    """
    Class to manage 'Parameter is missing' exception
    """

    def __init__(self, param_id: str):
        super().__init__('Parameter {} is missing'.format(param_id))

def generate_collection(params, param_ids: set):
    """
    Generate an array of parameters for the param_ids set
    """
    coll = []
    for param in params:
        if param.id in param_ids:
            coll.append(param)
    return coll

def reset_value_errors(asset) -> set:
    """
    Set the value errors of all parameters to '' if they were set
    """
    param_ids = set()
    for param in asset.params:
        if param.valueError:
            param.valueError = ''
            param_ids.add(param.id)
    return param_ids



AWS_ACCESS_KEY_PARAM = 'imco_aws_access_key_id'
AWS_SECRET_ACCESS_KEY_PARAM = 'imco_aws_secret_access_key'
AWS_CREDENTIAL_PARAMS = [AWS_ACCESS_KEY_PARAM, AWS_SECRET_ACCESS_KEY_PARAM]

DEPLOY_CLOUD_ACCOUNT = 'deploy_cloud_account'
IMCO_CUSTOMER_ID = 'imco_customer_id'
CONNECT_PARAM_CREDENTIALS_SOURCE = 'deploy_credentials_source'
CONNECT_PARAM_CREDENTIALS_SOURCE_INTEGRATED = 'repeat'
CONNECT_PARAM_CREDENTIALS_SOURCE_MANUAL = 'manual'

CONNECT_PARAM_CLOUD_ACCOUNT = 'deploy_cloud_account'
CONNECT_PARAM_IMCO_CUSTOMER_ID = 'imco_customer_id'



"""
The kaffeeservice API interface
"""
from datetime import datetime, timedelta
from typing import Tuple
from urllib.parse import quote, urlencode

import io
import secrets
import jwt

from requests import Session
from instantcoffee.kaffeeservice.exceptions import service_exception

class Api:
    """
    Kaffeeservice API
    """

    def __init__(self, vendor_id, logger, deployer_audience, deployer_endpoint, 
        deployer_jwt_key, ca_file=None):
        self._logger = logger
        self._logger.info('Initializing kaffeeservice API')
        self._vendor_id = vendor_id
        self._endpoint = deployer_endpoint.rstrip('/')
        self._audience = deployer_audience
        self._pem_key = deployer_jwt_key
        self._session = Session()
        if ca_file:
            self._session.verify = ca_file

    def set_headers(self) -> dict:
        """
        Set Authorization and Content-Type header
        """
        jwt_payload = {
            'iss': self._vendor_id,
            'aud': self._audience,
            'nbf': datetime.utcnow() - timedelta(seconds=10),
            'exp': datetime.utcnow() + timedelta(seconds=20),
            'jti': secrets.token_urlsafe(20)
        }
        token = jwt.encode(jwt_payload, self._pem_key, algorithm='PS512')
        return {
            'Content-Type': 'application/json',
            'Authorization': 'Bearer {}'.format(token)
        }


    def get(self, path: str, query: dict = None):
        """
        Sends a GET request

        :param path: relative path for request

        :return: json-decoded content of response

        :raises service_exception
        """
        if query:
            qstr = urlencode(query)
            path = '{}?{}'.format(path, qstr)
        self._logger.debug('GET {}{}'.format(self._endpoint, path))
        headers = self.set_headers()
        self._logger.debug(f'headers={headers}')
        response = self._session.get(self._endpoint + path, headers=headers)
        if not response:
            raise service_exception(response.status_code, response.content)
        self._logger.debug(f'response: {response.json()}')
        return response.json()

    def put(self, path: str, body):
        """
        Sends a PUT request

        :param path: path for the request
        :param body: arguments for the request

        :return: json-decoded content of response

        :raises service_exception
        """
        self._logger.debug('PUT {}{} with body {}'.format(self._endpoint, path, body))
        response = self._session.put(self._endpoint + path, json=body, headers=self.set_headers())
        if not response:
            raise service_exception(response.status_code, response.content)
        if response.content:
            self._logger.debug(f'response: {response.json()}')
            return response.json()
        self._logger.debug('No response')

    def post(self, path: str, body):
        """
        Sends a POST request

        :param path: path for the request
        :param body: arguments for the request

        :return: json-decoded content of response

        :raises service_exception
        """
        self._logger.debug('POST {}{} with body {}'.format(self._endpoint, path, body))
        headers = self.set_headers()
        self._logger.debug(f'headers={headers}')
        response = self._session.post(self._endpoint + path, json=body, headers=headers)
        if not response:
            raise service_exception(response.status_code, response.content)
        if response.content:
            self._logger.debug(f'response: {response.json()}')
            return response.json()
        self._logger.debug('No response')
        
    def delete(self, path: str) -> bool:
        """
        Sends a DELETE request

        :param path: path for the request

        Returns if delete was finished (False: only Accepted)

        :raises service_exception
        """
        self._logger.debug('DELETE {}{}'.format(self._endpoint, path))
        response = self._session.delete(self._endpoint + path, headers=self.set_headers())
        if not response:
            raise service_exception(response.status_code, response.content)
        if response.status_code == 202:
            return False
        return True

    def get_cat(self, hub_id: str, mp_id: str, name: str, version: str) -> str:
        """
        Calls ExistsCAT on the kaffeeservice.

        Returns cat_id

        Note: Do not use names and versions containing slashes. While we could easily encode them,
        kaffeeservice is currently not able to decode them as the decoding leads to arbitrary results
        """
        path = '/hubs/{}/marketplaces/{}/cats/{}/versions/{}'\
            .format(hub_id, mp_id, quote(name), quote(version))
        result = self.get(path)
        return result.get('id')

    def upload_cat(self, hub_id: str, mp_id: str, name: str, version: str, file: io.IOBase) -> str:
        """
        Upload CAT

        Data should be file-like object (but can also be bytes)
        
        :raises service_exception
        """
        path = '/hubs/{}/marketplaces/{}/cats/{}/versions/{}'\
            .format(hub_id, mp_id, quote(name), quote(version))
        self._logger.debug('POST {}{}'.format(self._endpoint, path))
        response = self._session.post(self._endpoint + path, headers=self.set_headers(), data=file)
        if not response:
            raise service_exception(response.status_code, response.content)
        self._logger.debug(response.json())
        result = response.json()
        return result.get('id')

    def create_customer(self, hub_id: str, mp_id: str, connect_id: str, connect_external_id: str, name: str) -> str:
        """
        Create customer, if it does not exist yet.

        Otherwise, get existing customer.

        Returns IMCO customer ID
        """
        path = '/hubs/{}/marketplaces/{}/customers'.format(hub_id, mp_id)
        body = {
            'id': connect_id,
            'external_id': connect_external_id,
            'name': name
        }
        return self.post(path, body).get('customer_id')

    def discover_integrated_accounts(self, hub_id: str, mp_id: str, customer_id: str, cloud_provider: str):
        """
        Launch discovery of integrated cloud accounts
        """
        path = '/hubs/{}/marketplaces/{}/customers/{}/integrated-accounts/discover'.format(hub_id, mp_id, customer_id)
        body = {'cloud_provider': cloud_provider}
        self.put(path, body)

    def list_integrated_accounts(self, hub_id: str, mp_id: str, customer_id: str, cloud_provider: str) -> dict:
        """
        Lists integrated accounts, if discovery is finished

        Returns a dict with finished: bool, and optionally accounts: List[str]
        Note: Failed discovery is handled as finished discovery
        """
        query = {'cloud_provider': cloud_provider}
        path = '/hubs/{}/marketplaces/{}/customers/{}/integrated-accounts'.\
                format(hub_id, mp_id, customer_id)
        return self.get(path, query)

    def create_cloud_account(self,
            hub_id: str, mp_id: str, customer_id: str, cloud_provider: str, credential_id: str, credentials: dict
        ) -> str:
        """
        Create a new stand-alone cloud account
        Reuse existing, if cloud account with same credential_id exists

        Returns name of cloud account
        """
        path = '/hubs/{}/marketplaces/{}/customers/{}/standalone-accounts'.format(hub_id, mp_id, customer_id)
        body = {
            'cloud_provider': cloud_provider,
            'credential_id': credential_id,
            'credentials': credentials
        }
        return self.post(path, body).get('cloud_account')

    def create_deployment(self,
            hub_id: str, mp_id: str, customer_id: str, label: str, vendor_cat_id: str, inputs: dict
        ) -> Tuple[str, str]:
        """
        Create a deployment with label

        Returns dict with cat_id (on customer level) and task_id

        If label already exists, we get a ValidationError with errors->name error list
        """
        path = '/hubs/{}/marketplaces/{}/customers/{}/deployments/{}'\
            .format(hub_id, mp_id, customer_id, quote(label))
        body = {
            'cat_id': vendor_cat_id,
            'inputs': inputs
        }

        result = self.post(path, body)
        return result.get('cat_id'), result.get('task_id')

    def get_deployment_task(self,
        hub_id: str, mp_id: str, customer_id: str, cat_id: str, task_id:str
    ) -> Tuple[bool, dict, str]:
        """
        Get deployment task

        Returns a dict with:
            - finished: bool (independently if success or failure)
            - outputs: dict (optional - successfully finished)
            - error_message: str (optional - failed)
        """
        path = '/hubs/{}/marketplaces/{}/customers/{}/cats/{}/tasks/{}'.\
                format(hub_id, mp_id, customer_id, cat_id, task_id)
        result = self.get(path)
        return result.get('finished'), result.get('outputs'), result.get('error_message')

    def undeploy(self, hub_id: str, mp_id: str, customer_id: str, label: str) -> bool:
        """
        Delete a deployment

        Returns if finished

        If label already exists, we get a ValidationError with errors->name error list
        """
        path = '/hubs/{}/marketplaces/{}/customers/{}/deployments/{}'\
            .format(hub_id, mp_id, customer_id, quote(label))

        return self.delete(path)



"""
This module implements kaffeeservice errors
"""
class ServiceException(Exception):
    """
   Service exception base
   """

    def __init__(self, status_code: int, d: str):
        self.status_code = status_code
        self.body = d
        super().__init__('Service exception: {} - Body: {}'.format(status_code, d))


class ConflictException(ServiceException):
    """
    Service Not Found Exception
    """

    def __init__(self, data: str):
        super().__init__(409, data)


class NotFoundException(ServiceException):
    """
    Service Not Found Exception
    """

    def __init__(self, data: str):
        super().__init__(404, data)


class ValidationException(ServiceException):
    """
    Service Validation Exception
    """

    def __init__(self, data: str):
        super().__init__(422, data)

EXCEPTIONS = {
    404: NotFoundException,
    409: ConflictException,
    422: ValidationException
}

def service_exception(status_code, data: bytes) -> ServiceException:
    """
    Get custom exception depending on status code

    :param status_code: http error code
    :param data: the exception data content

    :return: custom ServiceException
    """
    exc = EXCEPTIONS.get(status_code)
    if exc:
        return exc(data.decode('utf-8'))
    return ServiceException(status_code, data.decode('utf-8'))



"""
Validator of cloud credentials / cloud accounts

Searches integrated cloud accounts or alternatively checks credentials
"""
import datetime
import json

from abc import ABC, abstractmethod
from time import sleep

from instantcoffee.kaffeeservice import consts
from instantcoffee.kaffeeservice.api import Api
from instantcoffee.kaffeeservice.exceptions import ValidationException


class InvalidParamValue(Exception):
    """
    Exception to handle invalid values in dropdown parameters, etc.
    """

    def __init__(self, param_id: str, value: str):
        super().__init__('Invalid value {} for param {}'.format(value, param_id))

class Validator(ABC):
    """
    The credential validator validates the AWS / Azure credentials.

    Parameters:
    - req: the Connect request
    - kaffeeservice: the wrapper for the kaffeeservice microservice.
    - logger: The logger for the request

    It expects the following Connect parameters to be defined in the product:
    - 
    - CONNECT_PARAM_IMCO_CUSTOMER_ID
    - In case of "integrated" mode: CONNECT_PARAM_CREDENTIALS_SOURCE
      - ""/CONNECT_PARAM_CREDENTIALS_SOURCE_INTEGRATED: find integrated subscription
      - CONNECT_PARAM_CREDENTIALS_SOURCE_INTEGRATED: enter credentials manually
    """
    def __init__(self, req, kaffeeservice, logger):
        self._req= req
        self._kaffeeservice: Api = kaffeeservice
        self._logger= logger

    @abstractmethod
    def extract_credentials(self):
        """
        Extract cloud provider specific credentials from asset parameters
        """
        raise NotImplementedError # pragma: no cover

    @abstractmethod
    def mark_credentials_invalid(self, param_ids, logger, msg=None):
        """
        Mark cloud-provider specific credential parameters as invalid
        """
        raise NotImplementedError # pragma: no cover

    @abstractmethod
    def delete_secret_keys(self, param_ids: set):
        """
        Delete the cloud-provider specific secret key parameter
        """
        raise NotImplementedError # pragma: no cover

    @staticmethod
    @abstractmethod
    def cloud_provider_name() -> str:
        """
        Return the cloud provider name
        """
        raise NotImplementedError # pragma: no cover

    def validate(self, integrated: bool = True, inquiring: bool = True,
            timeout: datetime.datetime = None) -> set:
        """
        Validate cloud credentials in request

        Precondition: value errors have been cleared beforehand!

        Updates the parameter values or value errors inline.

        Returns a set of parameter ids (in Connect terminology: param names) that need to be updated
        """

        if not timeout:
            timeout = datetime.datetime.now()+datetime.timedelta(seconds=20)
        self._logger.info(f'Validate credentials - Timeout: {timeout}')
        asset = self._req.asset
        param_ids = set()

        customer_id = asset.get_param_value(consts.IMCO_CUSTOMER_ID)
        self._logger.debug(f'customer_id: {customer_id}')
        if not customer_id:
            self._logger.info('Create customer')
            customer = asset.tiers.customer
            value = self._kaffeeservice.create_customer(
                asset.connection.hub.id, asset.marketplace.id, customer.id, customer.external_id, customer.name)
            asset.update_parameter(consts.IMCO_CUSTOMER_ID, value)
            param_ids.add(consts.IMCO_CUSTOMER_ID)

        params = []
        cloud_account = asset.get_param_value(consts.DEPLOY_CLOUD_ACCOUNT)
        if cloud_account:
            # We assume that cloud provider did not change (currently guaranteed by Connect)
            self._logger.info(f'Validate: Cloud account already set to {cloud_account} - \
                nothing to do')
        elif integrated:
            self._logger.info('Try to get integrated subscription')
            credentials_source = asset.get_param_value(consts.CONNECT_PARAM_CREDENTIALS_SOURCE)
            self._logger.debug(f'credentials_source value: {credentials_source}')
            if not credentials_source or credentials_source == consts.CONNECT_PARAM_CREDENTIALS_SOURCE_INTEGRATED:
                self._find_integrated_account(param_ids, timeout=timeout)
            elif credentials_source == consts.CONNECT_PARAM_CREDENTIALS_SOURCE_MANUAL:
                params = self._validate_credentials(param_ids, inquiring=inquiring)
            else:
                raise InvalidParamValue(consts.CONNECT_PARAM_CREDENTIALS_SOURCE, credentials_source)
        else:
            self._logger.info('Standalone mode: Validate credentials')
            # Note: Don't allow cloud desk to get involved
            params = self._validate_credentials(param_ids, inquiring=inquiring)

        return param_ids, params

    def _find_integrated_account(self, param_ids: set, timeout: datetime.datetime):
        """
        Find integrated accounts.

        - If no integrated account found, or discovery timed out, will inquire credential_source parameter
        - If one integrated account found, will set this account as "cloud account param" and add id to param_ids
        - If more than one integrated account found, will set value error for cloud account param,
          setting value to first, and adding id to param_ids.
        """
        asset = self._req.asset
        customer_id = asset.get_param_value(consts.IMCO_CUSTOMER_ID)
        hub_id = asset.connection.hub.id
        mp_id = asset.marketplace.id
        cloud_provider = self.cloud_provider_name()
        self._kaffeeservice.discover_integrated_accounts(hub_id, mp_id, customer_id, cloud_provider)
        result = {'finished': False}
        self._logger.info(f'Seconds remaining: {(timeout - datetime.datetime.now()).seconds}')
        while not result.get('finished') and datetime.datetime.now() + datetime.timedelta(seconds=3) < timeout:
            sleep(3)
            result = self._kaffeeservice.list_integrated_accounts(hub_id, mp_id, customer_id, cloud_provider)
            self._logger.debug(f'result = {result}')
            self._logger.info(f'Seconds remaining: {(timeout - datetime.datetime.now()).seconds}')

        credentials_source = asset.get_param_value(consts.CONNECT_PARAM_CREDENTIALS_SOURCE)
        if result.get('finished'):
            cloud_accounts = result.get('cloud_accounts')
            if cloud_accounts:
                value_error = ''
                if len(cloud_accounts) > 1:
                    value_error = 'Please choose between {}'.format(', '.join(cloud_accounts))
                asset.update_parameter(consts.DEPLOY_CLOUD_ACCOUNT, cloud_accounts[0], value_error)
                param_ids.add(consts.DEPLOY_CLOUD_ACCOUNT)
            else:
                self._logger.info('Integrated cloud account not found')
                value_error = 'No {} subscription found'.format(cloud_provider)
                asset.update_parameter(consts.CONNECT_PARAM_CREDENTIALS_SOURCE, credentials_source, value_error)
                param_ids.add(consts.CONNECT_PARAM_CREDENTIALS_SOURCE)
        else:
            self._logger.info('Integrated cloud account discovery timed out')
            value_error = 'Discovery of {} subscription timed out'.format(cloud_provider)
            asset.update_parameter(consts.CONNECT_PARAM_CREDENTIALS_SOURCE, credentials_source, value_error)
            param_ids.add(consts.CONNECT_PARAM_CREDENTIALS_SOURCE)

    def _validate_credentials(self, param_ids: set, inquiring: bool):
        """
        Check if credentials have been entered manually, and validate them.

        If credentials have been entered and are valid, create cloud account,
        update the parameter and delete secret key.

        Otherwise, but only when inquiring (i.e. not in draft validation),
        mark credentials as missing or invalid (see CSO-1252).
        """
        self._logger.info('Get credentials from Connect params')
        cred_id, credentials = self.extract_credentials()
        self._logger.debug(f'cred_id: {cred_id}, credentials: {credentials}')
        if cred_id == "":
            self._logger.info('Credentials not given yet')
            # In draft validation, do not mark them as invalid. Don't trust CBC in handling credentials
            self._logger.debug(f'inquiring={inquiring}')
            if inquiring:
                return self.mark_credentials_invalid(param_ids, self._logger)
            return

        asset = self._req.asset
        customer_id = asset.get_param_value(consts.IMCO_CUSTOMER_ID)
        hub_id = asset.connection.hub.id
        mp_id = asset.marketplace.id
        try:
            self._logger.info('Try to create cloud account')
            cloud_account = self._kaffeeservice.create_cloud_account(
                hub_id, mp_id, customer_id, self.cloud_provider_name(), cred_id, credentials
            )
        except ValidationException as ex:
            self._logger.info('Credentials are invalid')
            # In draft validation, do not mark them as invalid. Don't trust CBC in handling credentials
            # In theory, we should not enter here during draft validation, anyway,
            # if the Connect param is well-defined as hidden
            if inquiring:
                data = json.loads(ex.body)
                msg = data.get('error')
                return self.mark_credentials_invalid(param_ids, self._logger, msg)
            return

        self._logger.info('Update deploy_cloud_account parameter')
        asset.update_parameter(consts.DEPLOY_CLOUD_ACCOUNT, cloud_account)
        param_ids.add(consts.DEPLOY_CLOUD_ACCOUNT)

        self._logger.info('Delete secret param')
        self.delete_secret_keys(param_ids)



"""
AWS validator "interacts" with Connect asset parameters related to AWS
"""
from typing import Dict, Tuple
from instantcoffee.kaffeeservice import consts
from instantcoffee.cloudcredentials.validation import Validator


class AwsValidator(Validator):
    """
    Check AWS credentials

    It assumes the following additional parameters to be present in the request:
    - imco_aws_access_key_id
    - imco_aws_secret_access_key
    """
    @staticmethod
    def cloud_provider_name() -> str:
        return 'AWS'

    def __init__(self, req, kaffeeservice, logger, doc_link):
        self._doc_link = doc_link
        super().__init__(req, kaffeeservice, logger)

    def extract_credentials(self) -> Tuple[str, Dict[str, str]]:
        """
        Extracts the AWS credentials, expected to exist as Connect parameters prefixed by imco_aws_
        """
        asset = self._req.asset
        credentials = {}
        for cred_param_id in consts.AWS_CREDENTIAL_PARAMS:
            cred_id = cred_param_id[len('imco_aws_'):]
            param = asset.get_param_value(cred_param_id)
            credentials[cred_id] = param.strip()

        return credentials['access_key_id'], credentials

    def mark_credentials_invalid(self, param_ids: set, logger,
        message: str = 'To learn how to obtain AWS credentials, please visit the following URL'):
        """
        Marks the AWS credentials fields as invalid (with message and doc link)
        """
        params = []
        for cred_param_id in consts.AWS_CREDENTIAL_PARAMS:
            param_val = self._req.asset.get_param_value(cred_param_id)
            if cred_param_id == consts.AWS_SECRET_ACCESS_KEY_PARAM:
                param_val = ''
            logger.debug(f'updating value error for param {cred_param_id}')
            value_error = '{}: {}'.format(message, self._doc_link)
            param_ids.add(cred_param_id)
            param = self._req.asset.update_parameter(cred_param_id, param_val, value_error)
            logger.debug(f'param to update: {param}')
            params.append(param)
        return params

    def delete_secret_keys(self, param_ids: set):
        """
        Delete the secret key parameter value
        """
        param_ids.add(consts.AWS_SECRET_ACCESS_KEY_PARAM)
        self._req.asset.update_parameter(consts.AWS_ACCESS_KEY_PARAM, 'no longer needed')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['instantcoffee',
 'instantcoffee.cloudcredentials',
 'instantcoffee.kaffeeservice']

package_data = \
{'': ['*']}

install_requires = \
['Faker',
 'connect-extension-runner>=25.0.0,<26.0.0',
 'cryptography>=37.0.2,<38.0.0',
 'extaasy>=0.1.0,<0.2.0',
 'mock>=4.0.3,<5.0.0',
 'pyjwt>=2.4.0,<3.0.0',
 'pytest>=7.1.2,<8.0.0',
 'responses>=0.20.0,<0.21.0']

setup_kwargs = {
    'name': 'instant-coffee',
    'version': '0.1.0',
    'description': 'Project description',
    'long_description': '# Instant Coffee\n\nA library for connecting to kaffeeservice in Instant Cloud Solutions\n',
    'author': 'Ingram Micro Inc.',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
